"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) /* istanbul ignore next */  if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    /* istanbul ignore next */  d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var utils_1 = require("../../utils/utils");
var gast_public_1 = require("./gast_public");
var interpreter_1 = require("./interpreter");
var rest_1 = require("./rest");
(function (PROD_TYPE) {
    PROD_TYPE[PROD_TYPE["OPTION"] = 0] = "OPTION";
    PROD_TYPE[PROD_TYPE["REPETITION"] = 1] = "REPETITION";
    PROD_TYPE[PROD_TYPE["REPETITION_MANDATORY"] = 2] = "REPETITION_MANDATORY";
    PROD_TYPE[PROD_TYPE["REPETITION_MANDATORY_WITH_SEPARATOR"] = 3] = "REPETITION_MANDATORY_WITH_SEPARATOR";
    PROD_TYPE[PROD_TYPE["REPETITION_WITH_SEPARATOR"] = 4] = "REPETITION_WITH_SEPARATOR";
    PROD_TYPE[PROD_TYPE["ALTERNATION"] = 5] = "ALTERNATION";
})(exports.PROD_TYPE || (exports.PROD_TYPE = {}));
var PROD_TYPE = exports.PROD_TYPE;
function buildLookaheadFuncForOr(occurrence, ruleGrammar, k, hasPredicates) {
    var lookAheadPaths = getLookaheadPathsForOr(occurrence, ruleGrammar, k);
    return buildAlternativesLookAheadFunc(lookAheadPaths, hasPredicates);
}
exports.buildLookaheadFuncForOr = buildLookaheadFuncForOr;
function buildLookaheadForTopLevel(rule, k) {
    var paths = interpreter_1.possiblePathsFrom(rule.definition, k);
    var lookAheadPaths = lookAheadSequenceFromAlternatives([paths]);
    return buildSingleAlternativeLookaheadFunction(lookAheadPaths[0]);
}
exports.buildLookaheadForTopLevel = buildLookaheadForTopLevel;
/**
 *  When dealing with an Optional production (OPTION/MANY/2nd iteration of AT_LEAST_ONE/...) we need to compare
 *  the lookahead "inside" the production and the lookahead immediately "after" it in the same top level rule (context free).
 *
 *  Example: given a production:
 *  ABC(DE)?DF
 *
 *  The optional '(DE)?' should only be entered if we see 'DE'. a single Token 'D' is not sufficient to distinguish between the two
 *  alternatives.
 *
 *  @returns A Lookahead function which will return true IFF the parser should parse the Optional production.
 */
function buildLookaheadFuncForOptionalProd(occurrence, ruleGrammar, prodType, k) {
    var lookAheadPaths = getLookaheadPathsForOptionalProd(occurrence, ruleGrammar, prodType, k);
    return buildSingleAlternativeLookaheadFunction(lookAheadPaths[0]);
}
exports.buildLookaheadFuncForOptionalProd = buildLookaheadFuncForOptionalProd;
function buildLookaheadForOption(optionOccurrence, ruleGrammar, k) {
    return buildLookaheadFuncForOptionalProd(optionOccurrence, ruleGrammar, PROD_TYPE.OPTION, k);
}
exports.buildLookaheadForOption = buildLookaheadForOption;
function buildLookaheadForMany(optionOccurrence, ruleGrammar, k) {
    return buildLookaheadFuncForOptionalProd(optionOccurrence, ruleGrammar, PROD_TYPE.REPETITION, k);
}
exports.buildLookaheadForMany = buildLookaheadForMany;
function buildLookaheadForManySep(optionOccurrence, ruleGrammar, k) {
    return buildLookaheadFuncForOptionalProd(optionOccurrence, ruleGrammar, PROD_TYPE.REPETITION_WITH_SEPARATOR, k);
}
exports.buildLookaheadForManySep = buildLookaheadForManySep;
function buildLookaheadForAtLeastOne(optionOccurrence, ruleGrammar, k) {
    return buildLookaheadFuncForOptionalProd(optionOccurrence, ruleGrammar, PROD_TYPE.REPETITION_MANDATORY, k);
}
exports.buildLookaheadForAtLeastOne = buildLookaheadForAtLeastOne;
function buildLookaheadForAtLeastOneSep(optionOccurrence, ruleGrammar, k) {
    return buildLookaheadFuncForOptionalProd(optionOccurrence, ruleGrammar, PROD_TYPE.REPETITION_MANDATORY_WITH_SEPARATOR, k);
}
exports.buildLookaheadForAtLeastOneSep = buildLookaheadForAtLeastOneSep;
/**
 * @param alternatives - a Sequence of possible paths for each alternative. example:
 *                       input:
 *                       [
 *                         [[A, B] [B]],  // alternative 1, with 2 paths
 *                         [[C]]            // alternative 2, with 1 Path
 *                         [[C, D, E], [C, D, F]] // // alternative 3, with 2 Path
 *                       ]
 *
 *                       output:
 *                       [
 *                         [[A] [B]],  // one Token is enough to identify alternative 1 so [A, B] --> [A]
 *                         [[C]]       // no changes for alternative 2
 *                         [[C, D]] // // the third Tokens (E/F) are not needed to identify this alternative
 *                       ]
 */
function lookAheadSequenceFromAlternatives(alternatives) {
    function isUniquePrefix(arr, item) {
        return utils_1.find(arr, function (currOtherPath) {
            return utils_1.every(item, function (currPathTok, idx) {
                return currPathTok === currOtherPath[idx];
            });
        }) === undefined;
    }
    var result = utils_1.map(alternatives, function (currAlt, currIdx) {
        var otherAlts = utils_1.reject(alternatives, function (curAltInner) { return curAltInner === currAlt; });
        var allOtherPaths = utils_1.reduce(otherAlts, function (result, currOtherAlt) {
            return result.concat(currOtherAlt);
        }, []);
        return utils_1.reduce(currAlt, function (currAltResult, currPath) {
            var minimalPath = [];
            for (var i = 0; i < currPath.length; i++) {
                minimalPath.push(currPath[i]);
                var isUnique = isUniquePrefix(allOtherPaths, minimalPath);
                if (isUnique) {
                    break;
                }
            }
            // At this point we either have a distinguishing path (unique) or a path the may be ambiguous
            // If it is a identical to a path from another alternative it is ambiguous, but if it is a strict prefix of a path from another
            // alternative than its ambiguous nature depends on the order of alternatives.
            if (!containsPath(currAltResult, minimalPath)) {
                // found one minimal path to distinguish this alternative.
                currAltResult.push(minimalPath);
                return currAltResult;
            }
            return currAltResult;
        }, []);
    });
    return result;
}
exports.lookAheadSequenceFromAlternatives = lookAheadSequenceFromAlternatives;
/**
 * @param alts
 * @param hasPredicates
 * @returns {function(): number}
 */
function buildAlternativesLookAheadFunc(alts, hasPredicates) {
    var numOfAlts = alts.length;
    // This version takes into account the predicates as well.
    if (hasPredicates) {
        /**
         * @returns {number} - The chosen alternative index
         */
        return function (orAlts) {
            // unfortunately the predicates must be extracted every single time
            // as they cannot be cached due to keep references to parameters(vars) which are no longer valid.
            // note that in the common case of no predicates, no cpu time will be wasted on this (see else block)
            var predicates = utils_1.map(orAlts, function (currAlt) { return currAlt.WHEN; });
            for (var t = 0; t < numOfAlts; t++) {
                var currAlt = alts[t];
                var currNumOfPaths = currAlt.length;
                var currPredicate = predicates[t];
                if (currPredicate && !currPredicate.call(this)) {
                    // if the predicate does not match there is no point in checking the paths
                    continue;
                }
                nextPath: for (var j = 0; j < currNumOfPaths; j++) {
                    var currPath = currAlt[j];
                    var currPathLength = currPath.length;
                    for (var i = 0; i < currPathLength; i++) {
                        var nextToken = this.LA(i + 1);
                        if (!(nextToken instanceof currPath[i])) {
                            // mismatch in current path
                            // try the next pth
                            continue nextPath;
                        }
                    }
                    // found a full path that matches.
                    // this will also work for an empty ALT as the loop will be skipped
                    return t;
                }
            }
            // none of the alternatives could be matched
            return -1;
        };
    }
    else {
        /**
         * @returns {number} - The chosen alternative index
         */
        return function () {
            for (var t = 0; t < numOfAlts; t++) {
                var currAlt = alts[t];
                var currNumOfPaths = currAlt.length;
                nextPath: for (var j = 0; j < currNumOfPaths; j++) {
                    var currPath = currAlt[j];
                    var currPathLength = currPath.length;
                    for (var i = 0; i < currPathLength; i++) {
                        var nextToken = this.LA(i + 1);
                        if (!(nextToken instanceof currPath[i])) {
                            // mismatch in current path
                            // try the next pth
                            continue nextPath;
                        }
                    }
                    // found a full path that matches.
                    // this will also work for an empty ALT as the loop will be skipped
                    return t;
                }
            }
            // none of the alternatives could be matched
            return -1;
        };
    }
}
exports.buildAlternativesLookAheadFunc = buildAlternativesLookAheadFunc;
function buildSingleAlternativeLookaheadFunction(alt) {
    // TODO: performance optimizations for the (common) edge case of K == 1
    var numOfPaths = alt.length;
    return function () {
        nextPath: for (var j = 0; j < numOfPaths; j++) {
            var currPath = alt[j];
            var currPathLength = currPath.length;
            for (var i = 0; i < currPathLength; i++) {
                var nextToken = this.LA(i + 1);
                if (!(nextToken instanceof currPath[i])) {
                    // mismatch in current path
                    // try the next pth
                    continue nextPath;
                }
            }
            // found a full path that matches.
            return true;
        }
        // none of the paths matched
        return false;
    };
}
exports.buildSingleAlternativeLookaheadFunction = buildSingleAlternativeLookaheadFunction;
var RestDefinitionFinderWalker = (function (_super) {
    __extends(RestDefinitionFinderWalker, _super);
    function RestDefinitionFinderWalker(topProd, targetOccurrence, targetProdType) {
        _super.call(this);
        this.topProd = topProd;
        this.targetOccurrence = targetOccurrence;
        this.targetProdType = targetProdType;
    }
    RestDefinitionFinderWalker.prototype.startWalking = function () {
        this.walk(this.topProd);
        return this.restDef;
    };
    RestDefinitionFinderWalker.prototype.checkIsTarget = function (node, expectedProdType, currRest, prevRest) {
        if (node.occurrenceInParent === this.targetOccurrence &&
            this.targetProdType === expectedProdType) {
            this.restDef = currRest.concat(prevRest);
            return true;
        }
        // performance optimization, do not iterate over the entire Grammar ast after we have found the target
        return false;
    };
    RestDefinitionFinderWalker.prototype.walkOption = function (optionProd, currRest, prevRest) {
        if (!this.checkIsTarget(optionProd, PROD_TYPE.OPTION, currRest, prevRest)) {
            _super.prototype.walkOption.call(this, optionProd, currRest, prevRest);
        }
    };
    RestDefinitionFinderWalker.prototype.walkAtLeastOne = function (atLeastOneProd, currRest, prevRest) {
        if (!this.checkIsTarget(atLeastOneProd, PROD_TYPE.REPETITION_MANDATORY, currRest, prevRest)) {
            _super.prototype.walkOption.call(this, atLeastOneProd, currRest, prevRest);
        }
    };
    RestDefinitionFinderWalker.prototype.walkAtLeastOneSep = function (atLeastOneSepProd, currRest, prevRest) {
        if (!this.checkIsTarget(atLeastOneSepProd, PROD_TYPE.REPETITION_MANDATORY_WITH_SEPARATOR, currRest, prevRest)) {
            _super.prototype.walkOption.call(this, atLeastOneSepProd, currRest, prevRest);
        }
    };
    RestDefinitionFinderWalker.prototype.walkMany = function (manyProd, currRest, prevRest) {
        if (!this.checkIsTarget(manyProd, PROD_TYPE.REPETITION, currRest, prevRest)) {
            _super.prototype.walkOption.call(this, manyProd, currRest, prevRest);
        }
    };
    RestDefinitionFinderWalker.prototype.walkManySep = function (manySepProd, currRest, prevRest) {
        if (!this.checkIsTarget(manySepProd, PROD_TYPE.REPETITION_WITH_SEPARATOR, currRest, prevRest)) {
            _super.prototype.walkOption.call(this, manySepProd, currRest, prevRest);
        }
    };
    return RestDefinitionFinderWalker;
}(rest_1.RestWalker));
/**
 * Returns the definition of a target production in a top level level rule.
 */
var InsideDefinitionFinderVisitor = (function (_super) {
    __extends(InsideDefinitionFinderVisitor, _super);
    function InsideDefinitionFinderVisitor(targetOccurrence, targetProdType) {
        _super.call(this);
        this.targetOccurrence = targetOccurrence;
        this.targetProdType = targetProdType;
        this.result = [];
    }
    InsideDefinitionFinderVisitor.prototype.checkIsTarget = function (node, expectedProdName) {
        if (node.occurrenceInParent === this.targetOccurrence &&
            this.targetProdType === expectedProdName) {
            this.result = node.definition;
        }
    };
    InsideDefinitionFinderVisitor.prototype.visitOption = function (node) {
        this.checkIsTarget(node, PROD_TYPE.OPTION);
    };
    InsideDefinitionFinderVisitor.prototype.visitRepetition = function (node) {
        this.checkIsTarget(node, PROD_TYPE.REPETITION);
    };
    InsideDefinitionFinderVisitor.prototype.visitRepetitionMandatory = function (node) {
        this.checkIsTarget(node, PROD_TYPE.REPETITION_MANDATORY);
    };
    InsideDefinitionFinderVisitor.prototype.visitRepetitionMandatoryWithSeparator = function (node) {
        this.checkIsTarget(node, PROD_TYPE.REPETITION_MANDATORY_WITH_SEPARATOR);
    };
    InsideDefinitionFinderVisitor.prototype.visitRepetitionWithSeparator = function (node) {
        this.checkIsTarget(node, PROD_TYPE.REPETITION_WITH_SEPARATOR);
    };
    InsideDefinitionFinderVisitor.prototype.visitAlternation = function (node) {
        this.checkIsTarget(node, PROD_TYPE.ALTERNATION);
    };
    return InsideDefinitionFinderVisitor;
}(gast_public_1.gast.GAstVisitor));
function getLookaheadPathsForOr(occurrence, ruleGrammar, k) {
    var visitor = new InsideDefinitionFinderVisitor(occurrence, PROD_TYPE.ALTERNATION);
    ruleGrammar.accept(visitor);
    var alternatives = utils_1.map(visitor.result, function (currAlt) { return interpreter_1.possiblePathsFrom([currAlt], k); });
    return lookAheadSequenceFromAlternatives(alternatives);
}
exports.getLookaheadPathsForOr = getLookaheadPathsForOr;
function getLookaheadPathsForOptionalProd(occurrence, ruleGrammar, prodType, k) {
    var insideDefVisitor = new InsideDefinitionFinderVisitor(occurrence, prodType);
    ruleGrammar.accept(insideDefVisitor);
    var insideDef = insideDefVisitor.result;
    var afterDefWalker = new RestDefinitionFinderWalker(ruleGrammar, occurrence, prodType);
    var afterDef = afterDefWalker.startWalking();
    var insidePaths = interpreter_1.possiblePathsFrom(insideDef, k);
    var afterPaths = interpreter_1.possiblePathsFrom(afterDef, k);
    return lookAheadSequenceFromAlternatives([insidePaths, afterPaths]);
}
exports.getLookaheadPathsForOptionalProd = getLookaheadPathsForOptionalProd;
function containsPath(alternative, path) {
    var found = utils_1.find(alternative, function (otherPath) {
        return path.length === otherPath.length &&
            utils_1.every(path, function (targetItem, idx) {
                return targetItem === otherPath[idx];
            });
    });
    return found !== undefined;
}
exports.containsPath = containsPath;
